// Gps.c
//
#define true 1
#define false 0

#define bool char

#define BUFFER_SIZE 1280
#define MSG_SIZE 128

#include "Resource.h"

// 1s delay
//
int cntTop;
int cnt10s;
char top;

char txt[128];

sbit Mmc_Chip_Select at LATD0_bit;
sbit Mmc_Chip_Select_Direction at TRISD0_bit;

// TFT Connections
//
char TFT_DataPort at PORTJ;
char TFT_DataPort_Direction at TRISJ;

sbit TFT_RST at LATD3_bit;
sbit TFT_BLED at LATC2_bit;
sbit TFT_RS at LATE0_bit;
sbit TFT_CS at LATD2_bit;
sbit TFT_RD at LATH1_bit;
sbit TFT_WR at LATH2_bit;
sbit TFT_RST_Direction at TRISD3_bit;
sbit TFT_BLED_Direction at TRISC2_bit;
sbit TFT_RS_Direction at TRISE0_bit;
sbit TFT_CS_Direction at TRISD2_bit;
sbit TFT_RD_Direction at TRISH1_bit;
sbit TFT_WR_Direction at TRISH2_bit;

// Touch Panel Connections
//
sbit DriveX_Left at LATA0_bit;
sbit DriveX_Right at LATD7_bit;
sbit DriveY_Up at LATA1_bit;
sbit DriveY_Down at LATD4_bit;
sbit DriveX_Left_Direction at TRISA0_bit;
sbit DriveX_Right_Direction at TRISD7_bit;
sbit DriveY_Up_Direction at TRISA1_bit;
sbit DriveY_Down_Direction at TRISD4_bit;

// UART Buffer
//
int iRead;
int iWrite;
int nBytes;
int nLF;
unsigned char buffer[BUFFER_SIZE];
unsigned char msg[MSG_SIZE+1];
bool bMsgReady;

unsigned int Xcoord, Ycoord;

int action;
enum ACTIONS { NOTHING=0, GPX};

const X_GPX_MIN = 129;
const X_GPX_MAX = 190;
const Y_GPX_MIN = 89;
const Y_GPX_MAX = 150;

const X_BAR = 298;
const Y_BAR = 16;

// Gpx file
//
bool bCapture;
char filename[] = "Test.gpx";

// GPS data
//
int intLat = 0;
int intLon = 0;
long decLat = 0;
long decLon = 0;

int nSatellites = 0;
int nSatellitesOld = -1;

char latitude[10];
char longitude[11];

char time[9];
char date[9];
char oldDate[9];

char nSats[3];
char altitude[8];

int nPars;
unsigned char* pPars[32];

unsigned char strApiSetDatum[] = "PMTK330,67";
unsigned char strApiSetDgpsMode[] = "PMTK301,1";
unsigned char strApiSetSbasEnabled[] = "PMTK313,1";

unsigned char strCmdFullColdStart[] = "PMTK104";

unsigned char strApiQDatum[] = "PMTK430";

// Timer 0 and UART Interrupt
//
void interrupt()
{
  int j;
  unsigned char cRead;
  
  // If message available and can be processed
  //
  if (( nLF > 0 ) && ( bMsgReady == false ))
  {
    j = 0;
    nLF--;
    bMsgReady = true;
    while ( 1 )
    {
      cRead = msg[j] = buffer[iRead];
      iRead++;
      nBytes--;
      if ( iRead == BUFFER_SIZE )
        iRead = 0;
      if ( bMsgReady == true )
      {
        j++;
        if ( j == MSG_SIZE )
          bMsgReady = false;
      }
      if ( cRead == 10 )
        break;
    }
  }
  
  // Timer ( 10mS )
  //
  if (TMR0IF_bit)
  {
    cntTop++;
    if ( cntTop > 100 )
    {
      // 1s delay
      top = 1;
      cntTop = 0;
    }
    TMR0H = 0x15;
    TMR0L = 0x9e;
    TMR0IF_bit = 0;
  }
  
  // UART Interrupt
  //
  if (RCIF_bit == 1)
  {
    cRead = UART1_Read();
    if ( nBytes == BUFFER_SIZE )
    {
      // Buffer is full, buffer is lost ( It's a bug !!! )
      //
      iRead = 0;
      iWrite = 0;
      nBytes = 0;
      nLF = 0;
      bMsgReady = false;
    }
    else
    {
      buffer[iWrite] = cRead;
      iWrite++;
      nBytes++;
      if ( iWrite == BUFFER_SIZE )
        iWrite = 0;
      // End of message seen ?
      if ( cRead == 10 )
        nLF++;
    }
    RCIF_bit = 0;
  }
}

// Initialize
//
void Init()
{

  PLLEN_bit = 1;
  Delay_ms(150);

  ANCON0 = 0xF0;
  ADSHR_bit  = 1;
  ANCON1  = 0xFF;
  ADSHR_bit  = 0;

  CM1CON  = 0;
  CM2CON  = 0;

  // Timer0 10mS
  //
  T0CON = 0x80;
  TMR0H = 0x15;
  TMR0L = 0x9e;
  TMR0IE_bit         = 1;

  TRISF7_bit = 0;
  LATF7_bit = 1;

  ADC_Init();
  TFT_Init(320, 240);
  TP_TFT_Init(320, 240, 0, 1);
  TP_TFT_Set_ADC_Threshold(900);
  TFT_Fill_Screen(CL_BLACK);

  RC1IE_bit = 1;
  GIE_bit = 1;
  PEIE_bit = 1;
  
}

// Calibrate Touch Panel
//
void Calibrate()
{
  TFT_Fill_Screen(CL_BLACK);
  TFT_Set_Pen(CL_WHITE, 3);
  TFT_Set_Font(TFT_defaultFont, CL_WHITE, FO_HORIZONTAL);
  TFT_Write_Text("Calibration touchez d'abord", 70, 100);
  TFT_Line(315, 1, 319, 1);
  TFT_Line(310, 10, 319, 1);
  TFT_Line(319, 5, 319, 1);
  TFT_Write_Text("ici", 290, 20);

  TP_TFT_Calibrate_Min();
  Delay_ms(500);

  TFT_Set_Pen(CL_BLACK, 3);
  TFT_Set_Font(TFT_defaultFont, CL_BLACK, FO_HORIZONTAL);
  TFT_Line(315, 1, 319, 1);
  TFT_Line(310, 10, 319, 1);
  TFT_Line(319, 5, 319, 1);
  TFT_Write_Text("ici", 290, 20);

  TFT_Set_Pen(CL_WHITE, 3);
  TFT_Set_Font(TFT_defaultFont, CL_WHITE, FO_HORIZONTAL);
  TFT_Line(0, 239, 0, 235);
  TFT_Line(0, 239, 5, 239);
  TFT_Line(0, 239, 10, 229);
  TFT_Write_Text("puis ici ", 15, 200);

  TP_TFT_Calibrate_Max();
  Delay_ms(500);

}

// Display Screen
//
void DisplayScreen()
{

  TFT_Set_Font(Tahoma12x15, CL_WHITE, FO_HORIZONTAL);

  TFT_Set_Brush(1, CL_BLACK, 0, LEFT_TO_RIGHT, CL_BLACK, CL_BLACK);
  TFT_Set_Pen(CL_BLACK, 1);
  TFT_Rectangle(0, 0, 319, 239);

  TFT_Image(X_GPX_MIN, Y_GPX_MIN, GPX_Start, 1);

}

// Decode a hex character
//
int DecoderHex(unsigned char hex)
{
  if ((hex >= '0' ) && ( hex <= '9'))
    return hex-'0';
  if ((hex >= 'A' ) && ( hex <= 'F'))
    return hex-'A'+10;
  if ((hex >= 'a' ) && ( hex <= 'f'))
    return hex-'a'+10;
    
  return 255;

}

// Process Longitude
//
void ProcessLongitude(unsigned char* pPar1, unsigned char* pPar2)
{
  int i;
  
  strncpy(longitude, pPar1, 10);
  if (strlen(longitude) != 10 )
    return;

  intLon = (longitude[0] - '0') * 100 + (longitude[1] - '0') * 10 + longitude[2] - '0';
  if(*pPar2 == 'W')
  {
    intLon = 0 - intLon;
  }

  decLon = 0;
  for (i=3;i<10;i++)
  {
    if ( longitude[i] != '.' )
    {
      decLon = decLon * 10 + longitude[i] - '0';
    }
  }

  decLon = decLon * 10;
  decLon = decLon / 6;

}

// Process Latitude
//
void ProcessLatitude(unsigned char* pPar1, unsigned char* pPar2)
{
  int i;
  
  strncpy(latitude, pPar1, 9);
  if (strlen(latitude) != 9 )
    return;

  intLat = (latitude[0] - '0') * 10 + latitude[1] - '0';
  if(*pPar2 == 'S')
  {
    intLon = 0 - intLon;
  }
  decLat = 0;
  for (i=2;i<9;i++)
  {
    if ( latitude[i] != '.' )
    {
      decLat = decLat * 10 + latitude[i] - '0';
    }
  }

  decLat = decLat * 10;
  decLat = decLat / 6;

}

// Process Time
//
void ProcessTime(unsigned char* pPar)
{
  if ( strlen(pPar) != 10 )
  {
    time[0] = 0;
  }
  else
  {
    strncpy(&time[0], &pPar[0], 2);
    strncpy(&time[3], &pPar[2], 2);
    strncpy(&time[6], &pPar[4], 2);
    time[2] = time[5] = ':';
  }
  
  time[sizeof(time)-1] = 0;
}

// Process Date
//
void ProcessDate(unsigned char* pPar)
{

  if ( strlen(pPar) != 6 )
  {
    date[0] = 0;
  }
  else
  {
    strncpy(&date[0], &pPar[0], 2);
    strncpy(&date[3], &pPar[2], 2);
    strncpy(&date[6], &pPar[4], 2);
    date[2] = date[5] = '/';
  }

  date[sizeof(date)-1] = 0;
}

// Process $GPRMC
//
void ProcessGPRMC()
{

  if ( nPars != 13 )
    return;
  
  if ( *pPars[2] != 'A' )
     return;

  ProcessLatitude(pPars[3], pPars[4]);
  ProcessLongitude(pPars[5], pPars[6]);
  ProcessTime(pPars[1]);
  ProcessDate(pPars[9]);
  
}

// Process $GPVTG
//
void ProcessGPVTG()
{

}

// Process $GPGGA
//
void ProcessGPGGA()
{

  if ( nPars != 15 )
    return;

  ProcessLatitude(pPars[2], pPars[3]);
  ProcessLongitude(pPars[4], pPars[5]);
  ProcessTime(pPars[1]);

  strncpy(nSats, pPars[7], 2);
  nSatellites = 0;
  if ((nSats[0] >= '0' ) && ( nSats[0] <= '9' ))
    nSatellites = nSats[0] - '0';
  if ((nSats[1] >= '0' ) && ( nSats[1] <= '9' ))
    nSatellites = nSatellites*10 + nSats[1] - '0';
  strncpy(altitude, pPars[9], 7);

}

// Process $GPGSA
//
void ProcessGPGSA()
{

}

// Process $GPGSV
//
void ProcessGPGSV()
{

}

// Process $GPGLL
//
void ProcessGPGLL()
{
  if ( nPars != 8 )
    return;

  if ( *pPars[6] != 'A' )
     return;

  ProcessLatitude(pPars[1], pPars[2]);
  ProcessLongitude(pPars[3], pPars[4]);
  ProcessTime(pPars[5]);

}

// Process $GPZDA
//
void ProcessGPZDA()
{

}

// Process $GPTXT
//
void ProcessGPTXT()
{

}

// Process GPS message
//
void ProcessGps()
{

  char tmp[64];

  int n;
  int nLen;
  int checkSum1;
  int checkSum2;
  unsigned char * pMsg;
  unsigned char c;
  
  // Search for LF
  //
  nLen = 0;
  pMsg = &msg;
  while ( *pMsg != 10 )
  {
    pMsg++;
    nLen++;
  }
  msg[nLen+1] = 0;
  if ( nLen < 5 )
    return;
   
  // Search for CR
  //
  nLen--;
  if ( msg[nLen] != 13 )
    return;
  
  // Decode Hex Checksum from message
  nLen--;
  c = msg[nLen];
  checkSum1 = DecoderHex(c);
  nLen--;
  checkSum1 += DecoderHex(msg[nLen])*16;
  nLen--;
  if ( msg[nLen] != '*' )
    return;
  nLen--;

  // Calculate checksum
  //
  checksum2 = 0;
  n = nLen;
  pMsg = &msg+1;
  while ( n > 0 )
  {
    checksum2 = checksum2 ^ *pMsg;
    pMsg++;
    n--;
  }
  
  // Checksum ok ?
  if ( checkSum1 != checkSum2 )
    return;
  
  // Split parameters ( split character is ',' or '*' )
  //
  pMsg = &msg;
  pPars[0] = pMsg;
  nPars = 1;
  while ( *pMsg != 13 )
  {
    if ( *pMsg == ',' )
    {
      *pMsg = 0;
      pPars[nPars++] = pMsg+1;
    }
    else if ( *pMsg == '*' )
    {
      *pMsg = 0;
      break;
    }
    *pMsg++;
  }

  // Process Message
  //
  if (strcmp(pPars[0], "$GPRMC" ) == 0 )
  {
    ProcessGPRMC();
  }
  else if (strcmp(pPars[0], "$GPVTG" ) == 0 )
  {
    ProcessGPVTG();
  }
  else if (strcmp(pPars[0], "$GPGGA" ) == 0 )
  {
    ProcessGPGGA();
  }
  else if (strcmp(pPars[0], "$GPGSA" ) == 0 )
  {
    ProcessGPGSA();
  }
  else if (strcmp(pPars[0], "$GPGSV" ) == 0 )
  {
    ProcessGPGSV();
  }
  else if (strcmp(pPars[0], "$GPGLL" ) == 0 )
  {
    ProcessGPGLL();
  }
  else if (strcmp(pPars[0], "$GPZDA" ) == 0 )
  {
    ProcessGPZDA();
  }
  else if (strcmp(pPars[0], "$GPTXT" ) == 0 )
  {
    ProcessGPTXT();
  }
  else if ( strcmp(pPars[0], "$PMTK501") == 0 )
  {
    TFT_Write_Text(pPars[1], 0, 120);
  }
}

// Send Message to GPS
//
void SendMessage(unsigned char * pMsg)
{
  int checksum = 0;
  int left, right;
  unsigned char * pTmp;
  
  // Calculate checksum
  //
  pTmp = pMsg;
  while ( *pTmp != 0)
  {
    checksum = checksum ^ *pTmp;
    pTmp++;
  }
  // Checksum to Hex
  //
  left = checksum / 16;
  right = checksum & 0x0f;
  if ( left < 10 )
    left += '0';
  else
    left += 'A' - 10;
  if ( right < 10 )
    right += '0';
  else
    right += 'A' - 10;

  // Send Message
  //
  UART1_Write('$');
  UART1_Write_Text(pMsg);

  UART1_Write('*');
  UART1_Write(left);
  UART1_Write(right);
  UART1_Write(13);
  UART1_Write(10);
  
}

// Query Datum
//
void ApiQDatum()
{
  SendMessage(strApiQDatum);
}

// Set Datum
//
void ApiSetDatum()
{
  SendMessage(strApiSetDatum);
}

// Set Dgps Mode
//
void ApiSetDgpsMode()
{
  SendMessage(strApiSetDgpsMode);
}

// Set Sbas
//
void ApiSetSbasEnabled()
{
  SendMessage(strApiSetSbasEnabled);
}

// Cmd Full Cold Start
//
void CmdFullColdStart()
{
  SendMessage(strCmdFullColdStart);
}

// Test commands
//
bool TestCommands()
{

  if ( TP_TFT_Press_Detect() == 1 )
  {
    if (TP_TFT_Get_Coordinates(&Xcoord, &Ycoord) == 0)
    {
      action = NOTHING;
      if ((Xcoord >= X_GPX_MIN) &&
          (Xcoord <= X_GPX_MAX) &&
          (Ycoord >= Y_GPX_MIN) &&
          (Ycoord <= Y_GPX_MAX))
      {
        action = GPX;
      }
    }
    return false;
  }

  if ( action > NOTHING )
    return true;

  return false;

}

// Main
//
void main(void)
{
  action = NOTHING;

  latitude[0] = 0;
  longitude[0] = 0;
  time[0] = 0;
  date[0] = 0;
  nSats[0] = 0;
  altitude[0] = 0;

  latitude[9] = 0;
  longitude[10] = 0;
  time[10] = 0;
  date[6] = 0;
  nSats[2] = 0;
  altitude[7] = 0;
  
  top = 0;
  cntTop = 0;

  // Init UART Buffer
  //
  iRead = 0;
  iWrite = 0;
  nBytes = 0;
  nLF = 0;
  bMsgReady = false;

  // Init GPX capture
  //
  bCapture = false;

  Init();
  Delay_100ms();
  
  Calibrate();
  
  DisplayScreen();

  UART1_Init(9600);

  CmdFullColdStart();
  Delay_100ms();

  ApiSetDatum();
  ApiSetDgpsMode();
  ApiSetSbasEnabled();

  ApiQDatum();
  
  while (1)
  {
    // Is there a message to process
    //
    if ( bMsgReady == true )
    {
      ProcessGps();
      bMsgReady = false;
    }
    
    // User entry
    //
    if ( TestCommands() == true )
    {
      if ( action == GPX )
      {
        if ( bCapture == false )
        {
          // Start Capture GPX
          //
          SPI1_Init_Advanced(_SPI_MASTER_OSC_DIV64, _SPI_DATA_SAMPLE_MIDDLE, _SPI_CLK_IDLE_LOW, _SPI_LOW_2_HIGH);

          if (Mmc_Fat_Init() == 0) 
          {
            SPI1_Init_Advanced(_SPI_MASTER_OSC_DIV4, _SPI_DATA_SAMPLE_MIDDLE, _SPI_CLK_IDLE_LOW, _SPI_LOW_2_HIGH);
            cnt10s = 9;
            
            Mmc_Fat_Assign(&filename, 0xA0);
            Mmc_Fat_Delete();
            
            Mmc_Fat_Set_File_Date(2010, 4, 19, 9, 0, 0);
            Mmc_Fat_Assign(&filename, 0xA0);
            Mmc_Fat_Rewrite();
            
            Mmc_Fat_Write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\r\n", 57);
            Mmc_Fat_Write("<gpx>\r\n", 7);
            Mmc_Fat_Write("  <trk>\r\n", 9);
            Mmc_Fat_Write("    <name>MikroE</name>\r\n", 25);
            Mmc_Fat_Write("    <trkseg>\r\n", 14);
            bCapture = true;
            TFT_Image(X_GPX_MIN, Y_GPX_MIN, GPX_Stop, 1);
          }
        }
        else
        {
          // Stop Capture GPX
          //
          Mmc_Fat_Write("    </trkseg>\r\n", 15);
          Mmc_Fat_Write("  </trk>\r\n", 10);
          Mmc_Fat_Write("</gpx>\r\n", 8);
          Mmc_Fat_Close();
          bCapture = false;
          TFT_Image(X_GPX_MIN, Y_GPX_MIN, GPX_Start, 1);
        }
      }
      action = NOTHING;
    }
    
    // Display data every 1s
    //
    if ( top == 1 )
    {
      if ( bCapture )
      {
        cnt10s++;
        if ( cnt10s == 10 )
        {
          cnt10s = 0;
          sprintf(txt, "      <trkpt lat=\"%d.%6.6ld\" lon=\"%d.%6.6ld\">\r\n", intLat, decLat, intLon, decLon);
          Mmc_Fat_Write(txt, strlen(txt));
          Mmc_Fat_Write("        <ele>", 13);
          Mmc_Fat_Write(altitude, sizeof(altitude));
          Mmc_Fat_Write("</ele>\r\n", 8);
          Mmc_Fat_Write("      </trkpt>\r\n", 16);
        }
      }
      
      if ( nSatellites != nSatellitesOld )
      {
        nSatellitesOld = nSatellites;
        if ( nSatellites < 2 )
        {
          TFT_Set_Pen(CL_GRAY, 4);
          TFT_V_Line(Y_BAR-4, Y_BAR, X_BAR);
        }
        else
        {
          TFT_Set_Pen(CL_BLUE, 4);
          TFT_V_Line(Y_BAR-4, Y_BAR, X_BAR);
        }
        if ( nSatellites < 4 )
        {
          TFT_Set_Pen(CL_GRAY, 4);
          TFT_V_Line(Y_BAR-8, Y_BAR, X_BAR+6);
        }
        else
        {
          TFT_Set_Pen(CL_BLUE, 4);
          TFT_V_Line(Y_BAR-8, Y_BAR, X_BAR+6);
        }     
        if ( nSatellites < 6 )
        {
          TFT_Set_Pen(CL_GRAY, 4);
          TFT_V_Line(Y_BAR-12, Y_BAR, X_BAR+12);
        }
        else
        {
          TFT_Set_Pen(CL_BLUE, 4);
          TFT_V_Line(Y_BAR-12, Y_BAR, X_BAR+12);
        }
        if ( nSatellites < 8 )
        {
          TFT_Set_Pen(CL_GRAY, 4);
          TFT_V_Line(Y_BAR-16, Y_BAR, X_BAR+18);
        }
        else
        {
          TFT_Set_Pen(CL_BLUE, 4);
          TFT_V_Line(Y_BAR-16, Y_BAR, X_BAR+18);
        }
        TFT_Set_Pen(CL_BLACK, 1);
      }
      if ( strcmp(date, oldDate) != 0 )
      {
        strcpy(oldDate, date);
        TFT_Rectangle(10, 210, 69, 225);
        TFT_Write_Text(date, 10, 210);
      }
      TFT_Rectangle(260, 210, 319, 225);
      TFT_Write_Text(time, 260, 210);
      top = 0;
      sprintf(txt, "Lat  :  %d.%6.6ld    Lon  :  %d.%6.6ld    Alt  :  %s", intLat, decLat, intLon, decLon, altitude);
      TFT_Rectangle(40, 32, 319, 47);
      TFT_Write_Text(txt, 40, 32);
    }
  }
}